# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/dst/trade/multiwoz/trade.py
https://github.com/thu-coai/ConvLab-2
'''
import copy
import re

import torch

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.collections.nlp.data.datasets.multiwoz_dataset.multiwoz_slot_trans import REF_SYS_DA
from nemo.collections.nlp.neural_types import *
from nemo.collections.nlp.utils.callback_utils import tensor2numpy
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

__all__ = ['TradeStateUpdateNM']


class TradeStateUpdateNM(NonTrainableNM):
    """
     Takes the predictions of the TRADE Dialogue state tracking model, 
     generates human-readable model output and updates the dialogue
     state with the TRADE predcitions
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'point_outputs_pred': NeuralType(('B', 'T', 'D', 'D'), LogitsType()),
            'gating_preds': NeuralType(('B', 'D', 'D'), LogitsType()),
            'belief_state': NeuralType(
                axes=[
                    AxisType(kind=AxisKind.Batch, is_list=True),
                    AxisType(kind=DialogAxisKind.Domain, is_list=True),  # 7 domains
                ],
                elements_type=MultiWOZBeliefState(),
            ),
            'user_uttr': NeuralType(axes=[AxisType(kind=AxisKind.Batch, is_list=True)], elements_type=Utterance()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            'belief_state': NeuralType(
                axes=[
                    AxisType(kind=AxisKind.Batch, is_list=True),
                    AxisType(kind=DialogAxisKind.Domain, is_list=True),  # 7 domains
                ],
                elements_type=MultiWOZBeliefState(),
            ),
            'request_state': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Sequence, is_list=True)],
                elements_type=StringType(),
            ),
        }

    def __init__(self, data_desc):
        """
        Initializes the object
        Args:
            data_desc (obj): data descriptor for MultiWOZ dataset, contains information about domains, slots, 
                and associated vocabulary
        """
        super().__init__()
        self.data_desc = data_desc

    def forward(self, gating_preds, point_outputs_pred, belief_state, user_uttr):
        """
        Processes the TRADE model output and updates the dialogue (belief) state with the model's predictions
        Args:
            user_uttr (str): user utterance
            request_state (dict): contains requestsed slots-slot_value pairs for each domain
            belief_state (dict): dialgoue belief state, containt slot-slot value pair for all domains
            gating_preds (float): TRADE model gating predictions
            point_outputs_pred (float): TRADE model pointers predictions
        Returns:
            updated request_state (dict)
            updated belief_state (dict)
        """
        gate_outputs_max, point_outputs_max = self.get_trade_prediction(gating_preds, point_outputs_pred)
        trade_output = self.get_human_readable_output(gate_outputs_max, point_outputs_max)[0]
        logging.debug('TRADE output: %s', trade_output)

        new_belief_state = self.reformat_belief_state(
            trade_output, copy.deepcopy(belief_state), self.data_desc.ontology_value_dict
        )
        # update request state based on the latest user utterance
        # extract current user output
        new_request_state = self.detect_requestable_slots(user_uttr.lower(), self.data_desc.det_dict)
        logging.debug('Belief State after TRADE: %s', belief_state)
        logging.debug('Request State after TRADE: %s', new_request_state)
        return new_belief_state, new_request_state

    def get_trade_prediction(self, gating_preds, point_outputs_pred):
        """
        Takes argmax of the model's predictions
        Args:
            gating_preds (float): TRADE model gating predictions
            point_outputs_pred (float): TRADE model output, contains predicted pointers
        Returns:
            gate_outputs_max_list (array): list of the gating predicions
            point_outputs_max_list (list of arrays): each array contains the pointers predictions
        """
        p_max = torch.argmax(point_outputs_pred, dim=-1)
        point_outputs_max = [tensor2numpy(p_max)]
        g_max = torch.argmax(gating_preds, axis=-1)
        gate_outputs_max = tensor2numpy(g_max)
        return gate_outputs_max, point_outputs_max

    def get_human_readable_output(self, gating_preds, point_outputs_pred):
        """
        Returns trade output in the human readable format
        Args:
            gating_preds (array): an array of gating predictions, TRADE model output
            point_outputs_pred (list of arrays): TRADE model output, contains predicted pointers
        Returns:
            output (list of strings): TRADE model output, each values represents domain-slot_name-slot_value, 
                for example, ['hotel-pricerange-cheap', 'hotel-type-hotel']
        """
        slots = self.data_desc.slots
        bi = 0
        predict_belief_bsz_ptr = []
        inverse_unpoint_slot = dict([(v, k) for k, v in self.data_desc.gating_dict.items()])

        for si, sg in enumerate(gating_preds[bi]):
            if sg == self.data_desc.gating_dict["none"]:
                continue
            elif sg == self.data_desc.gating_dict["ptr"]:
                pred = point_outputs_pred[0][0][si]

                pred = [self.data_desc.vocab.idx2word[x] for x in pred]

                st = []
                for e in pred:
                    if e == 'EOS':
                        break
                    else:
                        st.append(e)
                st = " ".join(st)
                if st == "none":
                    continue
                else:
                    predict_belief_bsz_ptr.append(slots[si] + "-" + str(st))
            else:
                predict_belief_bsz_ptr.append(slots[si] + "-" + inverse_unpoint_slot[sg])
        # predict_belief_bsz_ptr ['hotel-pricerange-cheap', 'hotel-type-hotel']
        output = [predict_belief_bsz_ptr]
        return output

    def reformat_belief_state(self, raw_state, bs, value_dict):
        '''
        Reformat TRADE model raw state into the default_state format
        Args:
            raw_state(list of strings): raw TRADE model output/state, each values represents domain-slot_name-slot_value,
                for example, ['hotel-pricerange-cheap', 'hotel-type-hotel']
            bs (dict): belief state - see nemo.collections.nlp.data.datasets.multiwoz.default_state for the format
            value_dict (dict): a dictionary of all slot values for MultiWOZ dataset
        Returns:
            bs (dict): reformatted belief state
        '''
        for item in raw_state:
            item = item.lower()
            slist = item.split('-', 2)
            domain = slist[0].strip()
            slot = slist[1].strip()
            value = slist[2].strip()
            if domain not in bs:
                raise Exception('Error: domain <{}> not in belief state'.format(domain))
            dbs = bs[domain]
            assert 'semi' in dbs
            assert 'book' in dbs
            slot = REF_SYS_DA[domain.capitalize()].get(slot, slot)
            # reformat some slots
            if slot == 'arriveby':
                slot = 'arriveBy'
            elif slot == 'leaveat':
                slot = 'leaveAt'
            if slot in dbs['semi']:
                dbs['semi'][slot] = self.normalize_value(value_dict, domain, slot, value)
            elif slot in dbs['book']:
                dbs['book'][slot] = value
            elif slot.lower() in dbs['book']:
                dbs['book'][slot.lower()] = value
            else:
                logging.warning(
                    'unknown slot name <{}> with value <{}> of domain <{}>\nitem: {}\n\n'.format(
                        slot, value, domain, item
                    )
                )
        return bs

    def normalize_value(self, value_set, domain, slot, value):
        """Normalized the value produced by NLU module to map it to the ontology value space.
        Args:
            value_set (dict): The value set of task ontology.
            domain (str): The domain of the slot-value pairs.
            slot (str): The slot of the value.
            value (str): The raw value detected by NLU module.
        Returns:
            value (str): The normalized value, which fits with the domain ontology.
        """
        slot = slot.lower()
        value = value.lower()
        value = ' '.join(value.split())
        try:
            assert domain in value_set
        except:
            raise Exception('domain <{}> not found in value set'.format(domain))
        if slot not in value_set[domain]:
            logging.warning('slot {} no in domain {}'.format(slot, domain))
            return value

        value_list = value_set[domain][slot]
        # exact match or containing match
        v = self._match_or_contain(value, value_list)
        if v is not None:
            return v
        # some transfomations
        cand_values = self._transform_value(value)
        for cv in cand_values:
            v = self._match_or_contain(cv, value_list)
            if v is not None:
                logging.warning('slot value found via _match_or_contain')
                return v
        # special value matching
        v = self.special_match(domain, slot, value)
        if v is not None:
            logging.warning('slot value found via special_match')
            return v
        logging.warning('Failed: domain {} slot {} value {}, raw value returned.'.format(domain, slot, value))
        return value

    def _transform_value(self, value):
        """makes clean up value transformations"""
        cand_list = []
        # a 's -> a's
        if " 's" in value:
            cand_list.append(value.replace(" 's", "'s"))
        # a - b -> a-b
        if " - " in value:
            cand_list.append(value.replace(" - ", "-"))
        # center <-> centre
        if value == 'center':
            cand_list.append('centre')
        elif value == 'centre':
            cand_list.append('center')
        # the + value
        if not value.startswith('the '):
            cand_list.append('the ' + value)
        return cand_list

    def minDistance(self, word1, word2):
        """The minimum edit distance between word 1 and 2."""
        if not word1:
            return len(word2 or '') or 0
        if not word2:
            return len(word1 or '') or 0
        size1 = len(word1)
        size2 = len(word2)
        tmp = list(range(size2 + 1))
        value = None
        for i in range(size1):
            tmp[0] = i + 1
            last = i
            for j in range(size2):
                if word1[i] == word2[j]:
                    value = last
                else:
                    value = 1 + min(last, tmp[j], tmp[j + 1])
                last = tmp[j + 1]
                tmp[j + 1] = value
        return value

    def _match_or_contain(self, value, value_list):
        """
        Matches value by exact match or containing
        Args:
            value (str): slot value
            value_list (list of str): list of possible slot_values
        Returns:
            matched value
        """
        if value in value_list:
            return value
        for v in value_list:
            if v in value or value in v:
                return v
        # fuzzy match, when len(value) is large and distance(v1, v2) is small
        for v in value_list:
            d = self.minDistance(value, v)
            if (d <= 2 and len(value) >= 10) or (d <= 3 and len(value) >= 15):
                return v
        return None

    def special_match(self, domain, slot, value):
        """special slot fuzzy matching"""
        matched_result = None
        if slot == 'arriveby' or slot == 'leaveat':
            matched_result = self._match_time(value)
        elif slot == 'price' or slot == 'entrance fee':
            matched_result = self._match_pound_price(value)
        elif slot == 'trainid':
            matched_result = self._match_trainid(value)
        elif slot == 'duration':
            matched_result = self._match_duration(value)
        return matched_result

    def _match_time(self, value):
        """Returns the time (leaveby, arriveat) in value, None if no time in value."""
        mat = re.search(r"(\d{1,2}:\d{1,2})", value)
        if mat is not None and len(mat.groups()) > 0:
            return mat.groups()[0]
        return None

    def _match_trainid(self, value):
        """Returns the trainID in value, None if no trainID."""
        mat = re.search(r"TR(\d{4})", value)
        if mat is not None and len(mat.groups()) > 0:
            return mat.groups()[0]
        return None

    def _match_pound_price(self, value):
        """Return the price with pounds in value, None if no trainID."""
        mat = re.search(r"(\d{1,2},\d{1,2} pounds)", value)
        if mat is not None and len(mat.groups()) > 0:
            return mat.groups()[0]
        mat = re.search(r"(\d{1,2} pounds)", value)
        if mat is not None and len(mat.groups()) > 0:
            return mat.groups()[0]
        if "1 pound" in value.lower():
            return '1 pound'
        if 'free' in value:
            return 'free'
        return None

    def _match_duration(self, value):
        """Return the durations (by minute) in value, None if no trainID."""
        mat = re.search(r"(\d{1,2} minutes)", value)
        if mat is not None and len(mat.groups()) > 0:
            return mat.groups()[0]
        return None

    def detect_requestable_slots(self, observation, det_dic):
        """
        Finds slot values in the observation (user utterance) and adds the to the requested slots list
        Args:
            observation (str): user utterance
            det_dic (dict):  a dictionary of slot_name + (slot_name_domain) value pairs from user dialogue acts
        Returns:
            result (dict): of the requested slots in a format: {domain: {slot_name}: 0}
        """
        result = {}
        observation = observation.lower()
        _observation = ' {} '.format(observation)
        for value in det_dic.keys():
            _value = ' {} '.format(value.strip())
            if _value in _observation:
                key, domain = det_dic[value].split('-')
                if domain not in result:
                    result[domain] = {}
                result[domain][key] = 0
        return result



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# =============================================================================
# Copyright 2019 Salesforce Research.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2
"""

import json
import os
import pickle
import random

from torch.utils.data import Dataset

from nemo.collections.nlp.data.datasets.multiwoz_dataset.multiwoz_slot_trans import REF_USR_DA
from nemo.utils import logging

__all__ = ['MultiWOZDataset', 'MultiWOZDataDesc']


class MultiWOZDataset(Dataset):
    """
    By default, use only vocab from training data
    Need to modify the code a little bit to create the vocab from all files
    """

    def __init__(self, data_dir, mode, domains, all_domains, vocab, gating_dict, slots, num_samples=-1, shuffle=False):

        logging.info(f'Processing {mode} data')
        self.data_dir = data_dir
        self.mode = mode
        self.gating_dict = gating_dict
        self.domains = domains
        self.all_domains = all_domains
        self.vocab = vocab
        self.slots = slots

        self.features, self.max_len = self.get_features(num_samples, shuffle)
        logging.info("Sample 0: " + str(self.features[0]))

    def get_features(self, num_samples, shuffle):
        if num_samples == 0:
            raise ValueError("num_samples has to be positive", num_samples)

        filename = f'{self.data_dir}/{self.mode}_dials.json'
        logging.info(f'Reading from {filename}')
        dialogs = json.load(open(filename, 'r'))

        domain_count = {}
        data = []
        max_resp_len, max_value_len = 0, 0

        for dialog_dict in dialogs:
            if num_samples > 0 and len(data) >= num_samples:
                break

            dialog_history = ""
            for domain in dialog_dict['domains']:
                if domain not in self.domains:
                    continue
                if domain not in domain_count:
                    domain_count[domain] = 0
                domain_count[domain] += 1

            for turn in dialog_dict['dialogue']:
                if num_samples > 0 and len(data) >= num_samples:
                    break

                turn_uttr = turn['system_transcript'] + ' ; ' + turn['transcript']
                turn_uttr_strip = turn_uttr.strip()
                dialog_history += turn["system_transcript"] + " ; " + turn["transcript"] + " ; "
                source_text = dialog_history.strip()

                turn_beliefs = fix_general_label_error_multiwoz(turn['belief_state'], self.slots)

                turn_belief_list = [f'{k}-{v}' for k, v in turn_beliefs.items()]

                gating_label, responses = [], []
                for slot in self.slots:
                    if slot in turn_beliefs:
                        responses.append(str(turn_beliefs[slot]))
                        if turn_beliefs[slot] == "dontcare":
                            gating_label.append(self.gating_dict["dontcare"])
                        elif turn_beliefs[slot] == "none":
                            gating_label.append(self.gating_dict["none"])
                        else:
                            gating_label.append(self.gating_dict["ptr"])
                    else:
                        responses.append("none")
                        gating_label.append(self.gating_dict["none"])

                sample = {
                    'ID': dialog_dict['dialogue_idx'],
                    'domains': dialog_dict['domains'],
                    'turn_domain': turn['domain'],
                    'turn_id': turn['turn_idx'],
                    'dialogue_history': source_text,
                    'turn_belief': turn_belief_list,
                    'gating_label': gating_label,
                    'turn_uttr': turn_uttr_strip,
                    'responses': responses,
                }

                sample['context_ids'] = self.vocab.tokens2ids(sample['dialogue_history'].split())
                sample['responses_ids'] = [
                    self.vocab.tokens2ids(y.split() + [self.vocab.eos]) for y in sample['responses']
                ]
                sample['turn_domain'] = self.all_domains[sample['turn_domain']]

                data.append(sample)

                resp_len = len(sample['dialogue_history'].split())
                max_resp_len = max(max_resp_len, resp_len)

        logging.info(f'Domain count{domain_count}')
        logging.info(f'Max response length{max_resp_len}')
        logging.info(f'Processing {len(data)} samples')

        if shuffle:
            logging.info(f'Shuffling samples.')
            random.shuffle(data)

        return data, max_resp_len

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        item = self.features[idx]
        return {
            'dialog_id': item['ID'],
            'turn_id': item['turn_id'],
            'turn_belief': item['turn_belief'],
            'gating_label': item['gating_label'],
            'context_ids': item['context_ids'],
            'turn_domain': item['turn_domain'],
            'responses_ids': item['responses_ids'],
        }


class Vocab:
    """
    Vocab class for MultiWOZ dataset
    UNK_token = 0
    PAD_token = 1
    SOS_token = 3
    EOS_token = 2
    """

    def __init__(self):
        self.word2idx = {'UNK': 0, 'PAD': 1, 'EOS': 2, 'BOS': 3}
        self.idx2word = ['UNK', 'PAD', 'EOS', 'BOS']
        self.unk_id = self.word2idx['UNK']
        self.pad_id = self.word2idx['PAD']
        self.eos_id = self.word2idx['EOS']
        self.bos_id = self.word2idx['BOS']
        self.unk, self.pad, self.eos, self.bos = 'UNK', 'PAD', 'EOS', 'BOS'

    def __len__(self):
        return len(self.idx2word)

    def add_word(self, word):
        if word not in self.word2idx:
            self.word2idx[word] = len(self.idx2word)
            self.idx2word.append(word)

    def add_words(self, sent, level):
        """
        level == 'utterance': sent is a string
        level == 'slot': sent is a list
        level == 'belief': sent is a dictionary
        """
        if level == 'utterance':
            for word in sent.split():
                self.add_word(word)
        elif level == 'slot':
            for slot in sent:
                domain, info = slot.split('-')
                self.add_word(domain)
                for subslot in info.split(' '):
                    self.add_word(subslot)
        elif level == 'belief':
            for slot, value in sent.items():
                domain, info = slot.split('-')
                self.add_word(domain)
                for subslot in info.split(' '):
                    self.add_word(subslot)
                for val in value.split(' '):
                    self.add_word(val)

    def tokens2ids(self, tokens):
        """Converts list of tokens to list of ids."""
        return [self.word2idx[w] if w in self.word2idx else self.unk_id for w in tokens]


class MultiWOZDataDesc:
    """
    Processes MultiWOZ dataset, creates vocabulary file and list of slots.
    """

    def __init__(self, data_dir, domains={"attraction": 0, "restaurant": 1, "taxi": 2, "train": 3, "hotel": 4}):
        logging.info(f'Processing MultiWOZ dataset')

        self.all_domains = {
            'attraction': 0,
            'restaurant': 1,
            'taxi': 2,
            'train': 3,
            'hotel': 4,
            'hospital': 5,
            'bus': 6,
            'police': 7,
        }
        self.gating_dict = {'ptr': 0, 'dontcare': 1, 'none': 2}

        self.data_dir = data_dir
        self.domains = domains
        self.vocab = Vocab()

        ontology_file = open(f'{self.data_dir}/ontology.json', 'r')
        self.ontology = json.load(ontology_file)

        # self.value_dict is ontology reformating in the following way, for example:
        # {'taxi-arrive by': list_of_values} -> {'taxi':{'arriveby': list_of_slot_values}}
        self.ontology_value_dict = json.load(open(f'{self.data_dir}/value_dict.json', 'r'))
        # detected dictionary of slot_name + (slot_name_domain) value pairs
        # from user dialogue acts
        self.det_dict = {}
        for domain, dic in REF_USR_DA.items():
            for key, value in dic.items():
                assert '-' not in key
                self.det_dict[key.lower()] = key + '-' + domain
                self.det_dict[value.lower()] = key + '-' + domain

        self.vocab_file = None
        self.slots = None

        self.get_slots()
        self.get_vocab()

    def get_vocab(self):
        self.vocab_file = f'{self.data_dir}/vocab.pkl'

        if os.path.exists(self.vocab_file):
            logging.info(f'Loading vocab from {self.data_dir}')
            self.vocab = pickle.load(open(self.vocab_file, 'rb'))
        else:
            self.create_vocab()

        logging.info(f'Vocab size {len(self.vocab)}')

    def get_slots(self):
        used_domains = [key for key in self.ontology if key.split('-')[0] in self.domains]
        self.slots = [k.replace(' ', '').lower() if 'book' not in k else k.lower() for k in used_domains]

    def create_vocab(self):
        self.vocab.add_words(self.slots, 'slot')

        filename = f'{self.data_dir}/train_dials.json'
        logging.info(f'Building vocab from {filename}')
        dialogs = json.load(open(filename, 'r'))

        max_value_len = 0

        for dialog_dict in dialogs:
            for turn in dialog_dict['dialogue']:
                self.vocab.add_words(turn['system_transcript'], 'utterance')
                self.vocab.add_words(turn['transcript'], 'utterance')

                turn_beliefs = fix_general_label_error_multiwoz(turn['belief_state'], self.slots)
                lengths = [len(turn_beliefs[slot]) for slot in self.slots if slot in turn_beliefs]
                lengths.append(max_value_len)
                max_value_len = max(lengths)

        logging.info(f'Saving vocab to {self.data_dir}')
        with open(self.vocab_file, 'wb') as handle:
            pickle.dump(self.vocab, handle)


def fix_general_label_error_multiwoz(labels, slots):
    label_dict = dict([label['slots'][0] for label in labels])
    GENERAL_TYPO = {
        # type
        "guesthouse": "guest house",
        "guesthouses": "guest house",
        "guest": "guest house",
        "mutiple sports": "multiple sports",
        "sports": "multiple sports",
        "mutliple sports": "multiple sports",
        "swimmingpool": "swimming pool",
        "concerthall": "concert hall",
        "concert": "concert hall",
        "pool": "swimming pool",
        "night club": "nightclub",
        "mus": "museum",
        "ol": "architecture",
        "colleges": "college",
        "coll": "college",
        "architectural": "architecture",
        "musuem": "museum",
        "churches": "church",
        # area
        "center": "centre",
        "center of town": "centre",
        "near city center": "centre",
        "in the north": "north",
        "cen": "centre",
        "east side": "east",
        "east area": "east",
        "west part of town": "west",
        "ce": "centre",
        "town center": "centre",
        "centre of cambridge": "centre",
        "city center": "centre",
        "the south": "south",
        "scentre": "centre",
        "town centre": "centre",
        "in town": "centre",
        "north part of town": "north",
        "centre of town": "centre",
        "cb30aq": "none",
        # price
        "mode": "moderate",
        "moderate -ly": "moderate",
        "mo": "moderate",
        # day
        "next friday": "friday",
        "monda": "monday",
        # parking
        "free parking": "free",
        # internet
        "free internet": "yes",
        # star
        "4 star": "4",
        "4 stars": "4",
        "0 star rarting": "none",
        # others
        "y": "yes",
        "any": "dontcare",
        "n": "no",
        "does not care": "dontcare",
        "not men": "none",
        "not": "none",
        "not mentioned": "none",
        '': "none",
        "not mendtioned": "none",
        "3 .": "3",
        "does not": "no",
        "fun": "none",
        "art": "none",
    }

    hotel_ranges = [
        "nigh",
        "moderate -ly priced",
        "bed and breakfast",
        "centre",
        "venetian",
        "intern",
        "a cheap -er hotel",
    ]
    locations = ["gastropub", "la raza", "galleria", "gallery", "science", "m"]
    detailed_hotels = ["hotel with free parking and free wifi", "4", "3 star hotel"]
    areas = ["stansted airport", "cambridge", "silver street"]
    attr_areas = ["norwich", "ely", "museum", "same area as hotel"]

    for slot in slots:
        if slot in label_dict.keys():
            # general typos
            if label_dict[slot] in GENERAL_TYPO.keys():
                label_dict[slot] = label_dict[slot].replace(label_dict[slot], GENERAL_TYPO[label_dict[slot]])

            # miss match slot and value
            if (
                (slot == "hotel-type" and label_dict[slot] in hotel_ranges)
                or (slot == "hotel-internet" and label_dict[slot] == "4")
                or (slot == "hotel-pricerange" and label_dict[slot] == "2")
                or (slot == "attraction-type" and label_dict[slot] in locations)
                or ("area" in slot and label_dict[slot] in ["moderate"])
                or ("day" in slot and label_dict[slot] == "t")
            ):
                label_dict[slot] = "none"
            elif slot == "hotel-type" and label_dict[slot] in detailed_hotels:
                label_dict[slot] = "hotel"
            elif slot == "hotel-star" and label_dict[slot] == "3 star hotel":
                label_dict[slot] = "3"
            elif "area" in slot:
                if label_dict[slot] == "no":
                    label_dict[slot] = "north"
                elif label_dict[slot] == "we":
                    label_dict[slot] = "west"
                elif label_dict[slot] == "cent":
                    label_dict[slot] = "centre"
            elif "day" in slot:
                if label_dict[slot] == "we":
                    label_dict[slot] = "wednesday"
                elif label_dict[slot] == "no":
                    label_dict[slot] = "none"
            elif "price" in slot and label_dict[slot] == "ch":
                label_dict[slot] = "cheap"
            elif "internet" in slot and label_dict[slot] == "free":
                label_dict[slot] = "yes"

            # some out-of-define classification slot values
            if (slot == "restaurant-area" and label_dict[slot] in areas) or (
                slot == "attraction-area" and label_dict[slot] in attr_areas
            ):
                label_dict[slot] = "none"

    return label_dict



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np
import torch

from nemo import logging

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']

GLOBAL_KEYS = [
    "loss",
    "per_example_loss",
    "beam_results",
    "src_ids",
    "src_first_tokens",
    "pred",
    "labels",
    "labels_mask",
]


def eval_iter_callback(tensors, global_vars, tokenizer):
    for key in GLOBAL_KEYS:
        if key not in global_vars.keys():
            global_vars[key] = []

    for kv, v in tensors.items():

        if "crossentropylossnm1" in kv:
            for per_example_loss in v:
                pel = per_example_loss.cpu().numpy().tolist()
                global_vars["per_example_loss"].extend(pel)

        if "logits" in kv:
            for pred in v:
                p = torch.argmax(pred, dim=-1).int().cpu().numpy().tolist()
                global_vars["pred"].extend(p)

        if "labels~" in kv:
            for label in v:
                l = label.cpu().numpy().tolist()
                global_vars["labels"].extend(l)

        if "labels_mask" in kv:
            for mask in v:
                m = mask.cpu().numpy().tolist()
                global_vars["labels_mask"].extend(m)


def eval_epochs_done_callback(global_vars, validation_dataset=None):
    losses = np.array(global_vars["per_example_loss"])
    eval_loss = np.mean(losses)
    global_vars["per_example_loss"] = []

    labels = np.array([np.array(n) for n in global_vars["labels"]])
    predictions = np.array([np.array(n) for n in global_vars["pred"]])
    labels_mask = np.array([np.array(n) for n in global_vars["labels_mask"]])
    for key in GLOBAL_KEYS:
        global_vars[key] = []

    lor = np.logical_or(labels == predictions, ~labels_mask.astype(np.bool))
    accuracy = np.mean(np.all(lor, axis=1).astype(np.float32))

    logging.info("------------------------------------------------------------")
    logging.info("Validation loss: {0}".format(np.round(eval_loss, 3)))
    logging.info("Sentence level accuracy: {0}".format(accuracy))
    logging.info("------------------------------------------------------------")

    return dict({"eval_loss": eval_loss})



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import random

import numpy as np

from nemo import logging
from nemo.collections.nlp.utils.callback_utils import (
    get_classification_report,
    get_f1_scores,
    list2str,
    plot_confusion_matrix,
    tensor2list,
)

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars):
    if "all_intent_preds" not in global_vars.keys():
        global_vars["all_intent_preds"] = []
    if "all_intent_labels" not in global_vars.keys():
        global_vars["all_intent_labels"] = []
    if "all_slot_preds" not in global_vars.keys():
        global_vars["all_slot_preds"] = []
    if "all_slot_labels" not in global_vars.keys():
        global_vars["all_slot_labels"] = []
    if "all_subtokens_mask" not in global_vars.keys():
        global_vars["all_subtokens_mask"] = []

    all_intent_logits, all_intent_labels = [], []
    all_slot_logits, all_slot_labels = [], []
    all_subtokens_mask = []
    for kv, v in tensors.items():
        if kv.startswith('intent_logits'):
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    all_intent_logits.append(tensor2list(logit_tensor))

        if kv.startswith('intents'):
            for v_tensor in v:
                for label_tensor in v_tensor:
                    all_intent_labels.append(tensor2list(label_tensor))

        if kv.startswith('slot_logits'):
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    all_slot_logits.append(tensor2list(logit_tensor))

        if kv.startswith('slots'):
            for v_tensor in v:
                for label_tensor in v_tensor:
                    all_slot_labels.extend(tensor2list(label_tensor))

        if kv.startswith('subtokens_mask'):
            for v_tensor in v:
                for subtokens_mask_tensor in v_tensor:
                    all_subtokens_mask.extend(tensor2list(subtokens_mask_tensor))

    all_intent_preds = list(np.argmax(np.asarray(all_intent_logits), 1))
    all_slot_preds = list(np.argmax(np.asarray(all_slot_logits), 2).flatten())
    global_vars["all_intent_preds"].extend(all_intent_preds)
    global_vars["all_intent_labels"].extend(all_intent_labels)
    global_vars["all_slot_preds"].extend(all_slot_preds)
    global_vars["all_slot_labels"].extend(all_slot_labels)
    global_vars["all_subtokens_mask"].extend(all_subtokens_mask)


def eval_epochs_done_callback(global_vars, intents_label_ids, slots_label_ids, graph_fold=None, normalize_cm=True):
    intent_labels = np.asarray(global_vars['all_intent_labels'])
    intent_preds = np.asarray(global_vars['all_intent_preds'])

    slot_labels = np.asarray(global_vars['all_slot_labels'])
    slot_preds = np.asarray(global_vars['all_slot_preds'])
    subtokens_mask = np.asarray(global_vars['all_subtokens_mask']) > 0.5

    slot_labels = slot_labels[subtokens_mask]
    slot_preds = slot_preds[subtokens_mask]

    # print predictions and labels for a small random subset of data
    sample_size = 20
    i = 0
    if intent_preds.shape[0] > sample_size + 1:
        i = random.randint(0, intent_preds.shape[0] - sample_size - 1)
    logging.info("Sampled i_preds: [%s]" % list2str(intent_preds[i : i + sample_size]))
    logging.info("Sampled intents: [%s]" % list2str(intent_labels[i : i + sample_size]))
    logging.info("Sampled s_preds: [%s]" % list2str(slot_preds[i : i + sample_size]))
    logging.info("Sampled slots: [%s]" % list2str(slot_labels[i : i + sample_size]))

    if graph_fold:
        # calculate, plot and save the confusion_matrix
        plot_confusion_matrix(
            intent_labels, intent_preds, graph_fold, intents_label_ids, normalize=normalize_cm, prefix='Intent'
        )
        plot_confusion_matrix(
            slot_labels, slot_preds, graph_fold, slots_label_ids, normalize=normalize_cm, prefix='Slot'
        )

    logging.info('Slot Prediction Results:')
    slot_accuracy = np.mean(slot_labels == slot_preds)
    logging.info(f'Slot Accuracy: {slot_accuracy}')
    f1_scores = get_f1_scores(slot_labels, slot_preds, average_modes=['weighted', 'macro', 'micro'])
    for k, v in f1_scores.items():
        logging.info(f'{k}: {v}')

    logging.info(f'\n {get_classification_report(slot_labels, slot_preds, label_ids=slots_label_ids)}')

    logging.info('Intent Prediction Results:')
    intent_accuracy = np.mean(intent_labels == intent_preds)
    logging.info(f'Intent Accuracy: {intent_accuracy}')
    f1_scores = get_f1_scores(intent_labels, intent_preds, average_modes=['weighted', 'macro', 'micro'])
    for k, v in f1_scores.items():
        logging.info(f'{k}: {v}')

    logging.info(f'\n {get_classification_report(intent_labels, intent_preds, label_ids=intents_label_ids)}')

    return dict({'intent_accuracy': intent_accuracy, 'slot_accuracy': slot_accuracy})



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data.datasets.lasertagger_dataset import LaserTaggerDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['LaserTaggerDataLayer']


class LaserTaggerDataLayer(TextDataLayer):
    """
    Data layer for LaserTagger from source (src) to target (tgt) editing tasks.

    Args:
        preprocessed_data (str): path to preprocessed train/validation/test data
        use_t2t_decoder (bool): whether to use Autoregressive Decoder
        dataset_type (Dataset):
                the underlying dataset. Default: LaserTaggerDataset
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids: indices of tokens which constitute batches of masked text segments
        input_mask: bool tensor with 0s in place of source tokens to be masked
        segment_ids: bool tensor with 0's and 1's to denote the text segment type
        tgt_ids: indices of target tokens which constitute batches of masked text segments
        labels_mask: bool tensor with 0s in place of label tokens to be masked
        labels: indices of tokens which should be predicted from each of the
            corresponding target tokens in tgt_ids
        loss_mask: bool tensor with 0s in place of label tokens to be masked used for
            CrossEntropyLossNM

        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "segment_ids": NeuralType(('B', 'T'), ChannelType()),
            "tgt_ids": NeuralType(('B', 'T'), LabelsType()),
            "labels_mask": NeuralType(('B', 'T'), MaskType()),
            "labels": NeuralType(('B', 'T'), LabelsType()),
            "loss_mask": NeuralType(('B', 'T'), MaskType()),
        }

    def __init__(
        self, preprocessed_data, use_t2t_decoder, batch_size, shuffle=False, dataset_type=LaserTaggerDataset,
    ):
        dataset_params = {
            'preprocessed_data': preprocessed_data,
            'use_t2t_decoder': use_t2t_decoder,
        }
        super().__init__(dataset_type, dataset_params, batch_size=batch_size, shuffle=shuffle)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import sentencepiece as spm

from nemo.collections.nlp.data.tokenizers.tokenizer_spec import TokenizerSpec

__all__ = ['SentencePieceTokenizer']


class SentencePieceTokenizer(TokenizerSpec):
    def __init__(self, model_path, special_tokens={}):
        self.tokenizer = spm.SentencePieceProcessor()
        self.tokenizer.Load(model_path)
        # without special tokens
        self.original_vocab_size = self.tokenizer.get_piece_size()
        self.vocab_size = self.tokenizer.get_piece_size()
        self.special_token_to_id = {}
        self.id_to_special_token = {}
        self.add_special_tokens(special_tokens)

    def text_to_tokens(self, text):
        tokens = []
        idx = 0
        last_idx = 0

        while 1:
            indices = {}

            for token in self.special_token_to_id:
                try:
                    indices[token] = text[idx:].index(token)
                except ValueError:
                    continue

            if len(indices) == 0:
                break

            next_token = min(indices, key=indices.get)
            next_idx = idx + indices[next_token]

            tokens.extend(self.tokenizer.encode_as_pieces(text[idx:next_idx]))
            tokens.append(next_token)
            idx = next_idx + len(next_token)

        tokens.extend(self.tokenizer.encode_as_pieces(text[idx:]))
        return tokens

    def text_to_ids(self, text):
        ids = []
        idx = 0
        last_idx = 0

        while 1:
            indices = {}

            for token in self.special_token_to_id:
                try:
                    indices[token] = text[idx:].index(token)
                except ValueError:
                    continue

            if len(indices) == 0:
                break

            next_token = min(indices, key=indices.get)
            next_idx = idx + indices[next_token]

            ids.extend(self.tokenizer.encode_as_ids(text[idx:next_idx]))
            ids.append(self.special_token_to_id[next_token])
            idx = next_idx + len(next_token)

        ids.extend(self.tokenizer.encode_as_ids(text[idx:]))
        return ids

    def tokens_to_text(self, tokens):
        return self.tokenizer.decode_pieces(tokens)

    def ids_to_text(self, ids):
        text = ""
        last_i = 0

        for i, id in enumerate(ids):
            if id in self.id_to_special_token:
                text += self.tokenizer.decode_ids(ids[last_i:i]) + " "
                text += self.id_to_special_token[id] + " "
                last_i = i + 1

        text += self.tokenizer.decode_ids(ids[last_i:])
        return text.strip()

    def tokens_to_ids(self, tokens):
        ids = []
        for token in tokens:
            ids.append(self.token_to_id(token))
        return ids

    def token_to_id(self, token):
        if token in self.special_token_to_id:
            return self.special_token_to_id[token]
        return self.tokenizer.piece_to_id(token)

    def ids_to_tokens(self, ids):
        tokens = []
        for id in ids:
            if id >= self.original_vocab_size:
                tokens.append(self.id_to_special_token[id])
            else:
                tokens.append(self.tokenizer.id_to_piece(id))
        return tokens

    def add_special_tokens(self, special_tokens):
        if isinstance(special_tokens, list):
            for token in special_tokens:
                if (
                    self.tokenizer.piece_to_id(token) == self.tokenizer.unk_id()
                    and token not in self.special_token_to_id
                ):
                    self.special_token_to_id[token] = self.vocab_size
                    self.id_to_special_token[self.vocab_size] = token
                    self.vocab_size += 1
        elif isinstance(special_tokens, dict):
            for token_name, token in special_tokens.items():
                setattr(self, token_name, token)
                if (
                    self.tokenizer.piece_to_id(token) == self.tokenizer.unk_id()
                    and token not in self.special_token_to_id
                ):
                    self.special_token_to_id[token] = self.vocab_size
                    self.id_to_special_token[self.vocab_size] = token
                    self.vocab_size += 1

    @property
    def pad_id(self):
        return self.tokens_to_ids([self.pad_token])[0]

    @property
    def bos_id(self):
        return self.tokens_to_ids([self.bos_token])[0]

    @property
    def eos_id(self):
        return self.tokens_to_ids([self.eos_token])[0]

    @property
    def sep_id(self):
        return self.tokens_to_ids([self.sep_token])[0]

    @property
    def cls_id(self):
        return self.tokens_to_ids([self.cls_token])[0]



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# __all__ = [
#     "NeMoCallback",
#     "SimpleLogger",
#     "TensorboardLogger",
#     "WandBLogger",
#     "CheckpointCallback",
#     "on_action_start",
#     "on_action_end",
#     "on_epoch_start",
#     "on_epoch_end",
#     "on_batch_start",
#     "on_batch_end",
#     "on_step_start",
#     "on_step_end",
# ]

import glob
import os
import time
from abc import ABC
from typing import Callable, List, Union

from nemo.core.deprecated_callbacks import (
    ActionCallback,
    EvaluatorCallback,
    ModuleSaverCallback,
    SimpleLossLoggerCallback,
    UnfreezeCallback,
    ValueSetterCallback,
    WandbCallback,
)
from nemo.core.neural_types import NmTensor
from nemo.utils import get_checkpoint_from_dir, logging
from nemo.utils.app_state import AppState

try:
    import wandb

    _WANDB_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    _WANDB_AVAILABLE = False


class NeMoCallback(ABC):
    """The base class for callbacks inside of NeMo. It contains no __init__ which children classes are responsible for.
    Each callback contains 8 functions which are called at different stages of train(). All functions must take as the
    first argument: the current action state. This state is a StateWrapper object.
    TODO: Add a link to documentation.
    """

    def on_action_start(self, state):
        pass

    def on_epoch_start(self, state):
        pass

    def on_batch_start(self, state):
        pass

    def on_step_start(self, state):
        pass

    def on_step_end(self, state):
        pass

    def on_batch_end(self, state):
        pass

    def on_epoch_end(self, state):
        pass

    def on_action_end(self, state):
        pass


def on_action_start(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_action_start callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_action_start(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_epoch_start(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_epoch_start callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_epoch_start(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_batch_start(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_batch_start callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_batch_start(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_step_start(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_step_start callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_step_start(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_step_end(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_step_end callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_step_end(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_batch_end(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_batch_end callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_batch_end(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_epoch_end(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_epoch_end callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_epoch_end(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


def on_action_end(func):
    """A function decorator that wraps a Callable inside the NeMoCallback object and runs the function with the
    on_action_end callback event.
    """

    class NeMoCallbackWrapper(NeMoCallback):
        def __init__(self, my_func):
            self._func = my_func

        def on_action_end(self, state):
            self._func(state)

    return NeMoCallbackWrapper(func)


class SimpleLogger(NeMoCallback):
    """A simple callback that prints tensors to screen. It's default option is to print the training loss every
    100 steps. Additional tensors can be printed by adding them to the tensors_to_log argument.

    args:
        step_freq (int): The frequency of printing to screen. Defaults to every 100 steps
        tensors_to_log (List of str or NmTensor): A list of either tensor names or NmTensors which will be printed
            every step_freq steps.
            Defaults to ["loss"] which only prints the loss.
    """

    def __init__(self, step_freq: int = 100, tensors_to_log: List[Union[str, NmTensor]] = ["loss"]):
        self.step_freq = step_freq
        self.tensors_to_log = tensors_to_log

    def on_step_end(self, state):
        if state["step"] % self.step_freq == 0:
            for tensor_key in self.tensors_to_log:
                tensor = state["tensors"].get_tensor(tensor_key)
                logging.info("%s: %s", tensor_key, tensor.data.cpu().numpy())


class TensorboardLogger(NeMoCallback):
    """A tensorboard callback that logs tensors using a tensorboard writer object. It's default option is to log
    the loss every 100 steps. Additional scalar tensors can be logged by adding them to the tensors_to_log
    argument. In order to log complex tensorboard entities, the custom_tb_log_func must be passed it. By default,
    it always logs the current epoch and the time taken per epoch.

    args:
        tb_writer (required): The tensorboard logger object.
        step_freq (int): The frequency of tensorboard logging. Defaults to every 100 steps
        tensors_to_log (List of str or NmTensor): A list of either tensor names or NmTensors which will be logged
            every step_freq steps.
            Defaults to ["loss"] which only prints the loss.
        custom_tb_log_func (func): custom_tb_log_func should accept three position arguments: tensors, tb_writer, and
            step. tensors is a list of pytorch tensors that correspond to the values of the NmTensors in
            tensors_to_log. tb_writer is the tensorboard logger passed to TensorboardLogger. step is the current
            step.
            Defaults to None which logs each tensors_to_log as a scalar.
        log_epoch (bool): Whether to log epoch and epoch training time to tensorboard.
            Defaults to True.
        log_lr (bool): Whether to log the learning rate to tensorboard.
            Defaults to True.
    """

    def __init__(
        self,
        tb_writer: 'torch.utils.tensorboard.SummaryWriter',
        step_freq: int = 100,
        tensors_to_log: List[Union[str, NmTensor]] = ["loss"],
        custom_tb_log_func: Callable[[Union[str, NmTensor]], None] = None,
        log_epoch: bool = True,
        log_lr: bool = True,
    ):
        self.step_freq = step_freq
        self.tensors_to_log = tensors_to_log
        self.tb_writer = tb_writer
        self.custom_tb_log_func = custom_tb_log_func
        self._last_epoch_start = None
        self._log_epoch = log_epoch
        self._log_lr = log_lr

    def on_epoch_start(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            self._last_epoch_start = time.time()

    def on_epoch_end(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            if self._log_epoch:
                epoch_time = time.time() - self._last_epoch_start
                self.tb_writer.add_scalar('misc/epoch', state["epoch"], state["step"])
                self.tb_writer.add_scalar('misc/epoch_time', epoch_time, state["step"])

    def on_step_end(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            if state["step"] % self.step_freq == 0:
                tensor_values = []
                for tensor_key in self.tensors_to_log:
                    tensor_values.append(state["tensors"].get_tensor(tensor_key))

                def tb_log_func(tensors, tb_writer, step):
                    for t, name in zip(tensors, self.tensors_to_log):
                        tb_writer.add_scalar(name, t, step)

                if self.custom_tb_log_func is not None:
                    tb_log_func = self.custom_tb_log_func
                tb_log_func(tensor_values, self.tb_writer, state["step"])

                if self._log_lr:
                    self.tb_writer.add_scalar('param/lr', state["optimizers"][0].param_groups[0]['lr'], state["step"])


class WandBLogger(NeMoCallback):
    """A [Weights & Biases](https://docs.wandb.com/) callback that logs tensors to W&B. It's default option is to
    log the loss every 100 steps. Additional scalar tensors can be logged by adding them to the tensors_to_log
    argument. By default, it always logs the current epoch and the time taken per epoch.

    args:
        step_freq (int): The frequency of Weights and Biases logging. Defaults to every 100 steps
        tensors_to_log (List of str or NmTensor): A list of either tensor names or NmTensors which will be logged
            every step_freq steps.
            Defaults to ["loss"] which only prints the loss.
        wandb_name(str): wandb experiment name.
            Defaults to None
        wandb_project(str): wandb project name.
            Defaults to None
        args: argparse flags which will be logged as hyperparameters.
            Defaults to None.
        log_epoch (bool): Whether to log epoch and epoch training time to Weights and Biases.
            Defaults to True.
        log_lr (bool): Whether to log epoch and epoch training time to Weights and Biases.
            Defaults to True.
    """

    def __init__(
        self,
        step_freq: int = 100,
        tensors_to_log: List[Union[str, NmTensor]] = ["loss"],
        wandb_name: str = None,
        wandb_project: str = None,
        args=None,
        log_epoch: bool = True,
        log_lr: bool = True,
    ):
        if not _WANDB_AVAILABLE:
            logging.error("Could not import wandb. Did you install it (pip install --upgrade wandb)?")
        self._step_freq = step_freq
        self._tensors_to_log = tensors_to_log
        self._name = wandb_name
        self._project = wandb_project
        self._args = args
        self._last_epoch_start = None
        self._log_epoch = log_epoch
        self._log_lr = log_lr

    def on_action_start(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            if _WANDB_AVAILABLE and wandb.run is None:
                wandb.init(name=self._name, project=self._project)
                if self._args is not None:
                    wandb.config.update(self._args)
            elif _WANDB_AVAILABLE and wandb.run is not None:
                logging.info("Re-using wandb session")
            else:
                logging.error("Could not import wandb. Did you install it (pip install --upgrade wandb)?")
                logging.info("Will not log data to weights and biases.")
                self._step_freq = -1

    def on_step_end(self, state):
        # log training metrics
        if state["global_rank"] is None or state["global_rank"] == 0:
            if state["step"] % self._step_freq == 0 and self._step_freq > 0:
                tensors_logged = {t: state["tensors"].get_tensor(t).cpu() for t in self._tensors_to_log}
                # Always log learning rate
                if self._log_lr:
                    tensors_logged['LR'] = state["optimizers"][0].param_groups[0]['lr']
                self._wandb_log(tensors_logged, state["step"])

    def on_epoch_start(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            self._last_epoch_start = time.time()

    def on_epoch_end(self, state):
        if state["global_rank"] is None or state["global_rank"] == 0:
            if self._log_epoch:
                epoch_time = time.time() - self._last_epoch_start
                self._wandb_log({"epoch": state["epoch"], "epoch_time": epoch_time}, state["step"])

    @staticmethod
    def _wandb_log(tensors_logged, step):
        if _WANDB_AVAILABLE:
            wandb.log(tensors_logged, step=step)


class CheckpointCallback(NeMoCallback):
    """A callback that does checkpointing of module weights and trainer (incl. optimizer) status.

    args:
        folder (str, required): A path where checkpoints are to be stored and loaded from if load_from_folder is
            None.
        load_from_folder (str): A path where checkpoints can be loaded from.
            Defaults to None.
        step_freq (int): How often in terms of steps to save checkpoints. One of step_freq or epoch_freq is
            required.
        epoch_freq (int): How often in terms of epochs to save checkpoints. One of step_freq or epoch_freq is
            required.
        checkpoints_to_keep (int): Number of most recent checkpoints to keep. Older checkpoints will be deleted.
            Defaults to 4.
        force_load (bool): Whether to crash if loading is unsuccessful.
            Defaults to False
    """

    def __init__(
        self,
        folder: str,
        load_from_folder: str = None,
        step_freq: int = -1,
        epoch_freq: int = -1,
        checkpoints_to_keep: int = 4,
        force_load: bool = False,
    ):
        if step_freq == -1 and epoch_freq == -1:
            logging.warning("No checkpoints will be saved because step_freq and epoch_freq are both -1.")

        if step_freq > -1 and epoch_freq > -1:
            logging.warning("You config the model to save by both steps and epochs. Please use one or the other")
            epoch_freq = -1

        self._step_freq = step_freq
        self._epoch_freq = epoch_freq
        self._folder = folder
        self._load_from_folder = load_from_folder if load_from_folder else folder
        self._ckpt2keep = checkpoints_to_keep
        self._saved_ckpts = []
        # If True, run will fail if we cannot load module weights
        self._force_load = force_load

    def __save_to(self, path, state):
        if state["global_rank"] is not None and state["global_rank"] != 0:
            return
        if not os.path.isdir(path):
            logging.info(f"Creating {path} folder")
            os.makedirs(path, exist_ok=True)
        unique_mod_names = set()
        for module in AppState().modules:
            if module.num_weights > 0:
                if str(module) in unique_mod_names:
                    raise NotImplementedError(
                        "There were two instances of the same module. Please overwrite __str__() of one of the "
                        "modules."
                    )
                unique_mod_names.add(str(module))
                if self._step_freq > -1:
                    filename = f"{module}-STEP-{state['step']}.pt"
                else:
                    filename = f"{module}-EPOCH-{state['epoch']}.pt"
                module.save_to(os.path.join(path, filename))

        if self._step_freq > -1:
            filename = f"trainer-STEP-{state['step']}.pt"
            state.save_state_to(f"{path}/{filename}")
            self._saved_ckpts.append(f"-{state['step']}.pt")
        else:
            filename = f"trainer-EPOCH-{state['epoch']}.pt"
            state.save_state_to(f"{path}/{filename}")
            self._saved_ckpts.append(f"-{state['epoch']}.pt")

        if len(self._saved_ckpts) > self._ckpt2keep:
            for end in self._saved_ckpts[: -self._ckpt2keep]:
                for file in glob.glob(f'{path}/*{end}'):
                    os.remove(file)
            self._saved_ckpts = self._saved_ckpts[-self._ckpt2keep :]
        logging.info(f'Saved checkpoint: {path}/{filename}')

    def __restore_from(self, path, state):
        if not os.path.isdir(path):
            if self._force_load:
                raise ValueError("force_load was set to True for checkpoint callback but a checkpoint was not found.")
            logging.warning(f"Checkpoint folder {path} not found!")
        else:
            logging.info(f"Found checkpoint folder {path}. Will attempt to restore checkpoints from it.")
            modules_to_restore = []
            modules_to_restore_name = []
            for module in AppState().modules:
                if module.num_weights > 0:
                    modules_to_restore.append(module)
                    modules_to_restore_name.append(str(module))
            step_check = None
            try:
                module_checkpoints, steps = get_checkpoint_from_dir(modules_to_restore_name, path, return_steps=True)

                # If the steps are different, print a warning message
                for step in steps:
                    if step_check is None:
                        step_check = step
                    elif step != step_check:
                        logging.warning("Restoring from modules checkpoints where the training step does not match")
                        break

                for mod, checkpoint in zip(modules_to_restore, module_checkpoints):
                    mod.restore_from(checkpoint, state["local_rank"])
            except (ValueError) as e:
                if self._force_load:
                    raise ValueError(
                        "force_load was set to True for checkpoint callback but a checkpoint was not found."
                    )
                logging.warning(e)
                logging.warning(
                    f"Checkpoint folder {path} was present but nothing was restored. Continuing training from random "
                    "initialization."
                )
                return

            try:
                trainer_checkpoints, steps = get_checkpoint_from_dir(["trainer"], path, return_steps=True)
                if step_check is not None and step_check != steps[0]:
                    logging.error(
                        "The step we are restoring from the trainer checkpoint does not match one or more steps that "
                        "are being restored from modules."
                    )
                state.restore_state_from(trainer_checkpoints[0])
            except (ValueError) as e:
                logging.warning(e)
                logging.warning(
                    "Trainer state such as optimizer state and current step/epoch was not restored. Pretrained weights"
                    " have still been restore and fine-tuning should continue fine."
                )
                return

    def on_action_start(self, state):
        num_parameters = 0
        unique_mod_names = set()
        for module in AppState().modules:
            if module.num_weights > 0:
                if str(module) in unique_mod_names:
                    raise NotImplementedError(
                        "There were two instances of the same module. Please overwrite __str__() of one of the "
                        "modules."
                    )
                unique_mod_names.add(str(module))
                num_parameters += module.num_weights
        logging.info(f"Found {len(unique_mod_names)} modules with weights:")
        for name in unique_mod_names:
            logging.info(f"{name}")
        logging.info(f"Total model parameters: {num_parameters}")
        self.__restore_from(self._load_from_folder, state)

    def on_step_end(self, state):
        step = state["step"]
        if self._step_freq > 0 and step % self._step_freq == 0 and step > 0:
            self.__save_to(self._folder, state)

    def on_action_end(self, state):
        if self._step_freq > 0 or self._epoch_freq > 0:
            self.__save_to(self._folder, state)

    def on_epoch_end(self, state):
        epoch = state["epoch"]
        if self._epoch_freq > 0 and epoch % self._epoch_freq == 0 and epoch > 0:
            self.__save_to(self._folder, state)



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import time
import warnings
from collections import OrderedDict

import numpy as np
import onnx
import tensorrt as trt

from .tensorrt_format import FormatManager
from .tensorrt_runner import (
    DEFAULT_SHAPE_VALUE,
    TRT_LOGGER,
    TensorRTRunnerV2,
    default_value,
    find_in_dict,
    get_input_metadata_from_profile,
    is_dimension_dynamic,
    is_shape_dynamic,
    is_valid_shape_override,
    send_on_queue,
    write_timestamped,
)
from nemo import logging, logging_mode


def set_onnx_logging_level(sev):
    if sev >= logging.INFO:
        warnings.filterwarnings("ignore")


class BaseDataLoader(object):
    """
    Responsible for fetching or generting input data for runners.
    """

    def __call__(self, index, input_metadata, input_example=None):
        """
        Fetches or generates inputs.

        Args:
            index (int): The index of inputs to fetch. For any given index, the inputs should always be the same.
            input_metadata (OrderedDict[str, Tuple[np.dtype, Tuple[int]]]): Mapping of input names to their data types and shapes.

        Returns:
            OrderedDict[str, np.ndarray]: Mapping of input names to numpy buffers containing data.
        """
        raise NotImplementedError("BaseDataLoader is an abstract class")


class DefaultDataLoader(BaseDataLoader):
    def __init__(
        self,
        seed=None,
        default_shape_value=None,
        default_shapes=None,
        int_min=None,
        int_max=None,
        float_min=None,
        float_max=None,
    ):
        """
        Optional Args:
            seed (int): The seed to use when generating random inputs.
            default_shape_value (int): The default value to use when a dimension is dynamic.
            default_shapes (Dict[str, Tuple[int]]): A mapping of input names to their corresponding shapes.
        """
        self.seed = default_value(seed, int(time.time()))
        self.default_shapes = default_value(default_shapes, {})
        self.default_shape_value = default_value(default_shape_value, DEFAULT_SHAPE_VALUE)
        self.int_min = default_value(int_min, 1)
        self.int_max = default_value(int_max, 25)
        self.float_min = default_value(float_min, -1.0)
        self.float_max = default_value(float_max, 1.0)

    def __call__(self, index, input_metadata, input_example=None):
        logging.debug("Updating seed to: {:}".format(self.seed + index))
        rng = np.random.RandomState(self.seed + index)

        buffers = OrderedDict()
        i = 0
        for name, (dtype, shape) in input_metadata.items():
            if input_example is not None and (not isinstance(input_example, tuple) or i < len(input_example)):
                if isinstance(input_example, tuple):
                    static_shape = input_example[i].shape
                elif isinstance(input_example, OrderedDict):
                    static_shape = tuple(input_example.values())[i].shape
                else:
                    static_shape = [tuple(input_example.shape)]
            elif is_shape_dynamic(shape):
                if name in self.default_shapes:
                    static_shape = self.default_shapes[name]
                else:
                    static_shape = [self.default_shape_value if is_dimension_dynamic(elem) else elem for elem in shape]
                if static_shape != shape:
                    if not is_valid_shape_override(static_shape, shape):
                        logging.critical(
                            "Cannot override original shape: {:}, for input: {:} to {:}".format(
                                shape, name, static_shape
                            )
                        )
                    logging.warning(
                        "Input: {:}: Adjusted dynamic shape: {:} to: {:}".format(name, shape, static_shape),
                        mode=logging_mode.ONCE,
                    )
            else:
                if name in self.default_shapes:
                    logging.warning(
                        "Will not override static shape: {:}, for input: {:}".format(shape, name),
                        mode=logging_mode.ONCE,
                    )
                static_shape = shape

            if input_example is not None and (not isinstance(input_example, tuple) or i < len(input_example)):
                if isinstance(input_example, OrderedDict):
                    buffers[name] = list(input_example.values())[i].cpu()
                else:
                    buffers[name] = input_example[i].cpu() if isinstance(input_example, tuple) else input_example.cpu()
            elif np.issubdtype(dtype, np.integer):
                buffers[name] = rng.randint(low=self.int_min, high=self.int_max, size=static_shape, dtype=dtype)
            elif np.issubdtype(dtype, np.bool_):
                buffers[name] = rng.randint(low=0, high=2, size=static_shape).astype(dtype)
            else:
                buffers[name] = (
                    rng.random_sample(size=static_shape) * (self.float_max - self.float_min) + self.float_min
                ).astype(dtype)

            buffers[name] = np.array(
                buffers[name]
            )  # To handle scalars. The above functions return a float if shape is ().

            # If the shape is 1D, and has a length equal to the rank of the provided default shape, it is
            # likely to be a TRT shape tensor, and so should be overriden such that it's value (not shape) is the default shape.
            is_shape_tensor = (
                (not is_shape_dynamic(shape))
                and (name in self.default_shapes)
                and (len(shape) == 1)
                and (shape[0] == len(self.default_shapes[name]))
            )
            if is_shape_tensor:
                buffers[name] = np.array(self.default_shapes[name], dtype=dtype)
                logging.warning(
                    "Assuming {:} is a shape tensor. Setting to: {:}".format(name, buffers[name]),
                    mode=logging_mode.ONCE,
                )
            i = i + 1

        return buffers


# Caches data loaded by a DataLoader for use across multiple runners.
class DataLoaderCache(object):
    def __init__(self, data_loader):
        self.data_loader = data_loader
        self.cache = {}  # Dict[int, OrderedDict[str, np.ndarray]]

    def load(self, iteration, input_metadata, input_example=None):
        """
        Load the specified iteration from the cache if present, or generate using the data loader.

        Args:
            iteration (int): The iteration whose data to retrieve.
            input_metadata (OrderedDict[str, Tuple[np.dtype, Tuple[int]]]): Input Metadata, including shape and type information. The loader may attempt to match input_metadata when data in the cache does not exactly match a new set of input_metadata.
        """
        if iteration not in self.cache:
            logging.debug("Iteration {:} not found in cache, generating new buffers for all inputs".format(iteration))
            self.cache[iteration] = self.data_loader(iteration, input_metadata, input_example)
            if self.cache[iteration] is None:
                logging.critical(
                    "Received no data from data_loader(iteration, input_metadata) for input_metadata: {:}".format(
                        input_metadata
                    )
                )
        else:
            logging.info("Found iteration {:} in cache".format(iteration))

        feed_dict = OrderedDict()
        for index, (name, (dtype, shape)) in enumerate(input_metadata.items()):
            cached_name = find_in_dict(name, self.cache[iteration], index)
            if cached_name is None:
                logging.warning("Could not find input: {:} in cache, regenerating buffers".format(name))
                self.cache[iteration] = self.data_loader(iteration, input_metadata, input_example)
                cached_name = name

            buffer = self.cache[iteration][cached_name]

            if dtype != buffer.dtype:
                logging.warning(
                    "Cached buffer data type does not match data type for input: {:}. Note: Cached type: {:}, input type: {:}. Attempting to cast".format(
                        name, buffer.dtype, dtype
                    )
                )
                buffer = buffer.astype(dtype)

            if not is_valid_shape_override(buffer.shape, shape):
                logging.warning(
                    "Cached buffer shape does not match shape for input. Note: Cached shape: {:}, input shape: {:}.".format(
                        buffer.shape, shape
                    )
                )
                # Try to permute the shape to match
                try:
                    perm = FormatManager.permutation(
                        FormatManager.deduce_format(buffer.shape), FormatManager.deduce_format(shape)
                    )
                    new_shape = FormatManager.convert(tuple(buffer.shape), FormatManager.deduce_format(shape))
                    logging.warning(
                        "Attempting to permute shape: {:} using permutation {:}. New shape: {:}".format(
                            buffer.shape, perm, new_shape
                        )
                    )
                    buffer = np.transpose(buffer, perm)
                except NotImplementedError as err:
                    # If the FormatManager does not recognize the format, skip permutation.
                    logging.info("Skipping permutation due to {:}".format(err))
                except KeyError as err:
                    # If the FormatManager cannot generate the permutation for the format combination, skip permutation.
                    logging.info("Skipping permutation due to {:}".format(err))

            feed_dict[name] = buffer
        return feed_dict


class BaseModelLoader(object):
    """
    Loads a model for a runner.
    """

    def __call__(self):
        """
        Load the model.

        Returns:
            A model usable by the runner. The return type is dependent on the runner the loader has been implemented for.
        """
        raise NotImplementedError("BaseModelLoader is an abstract class")


class BaseOnnxModelLoader(BaseModelLoader):
    def check(self, model):
        try:
            onnx.checker.check_model(model)
            logging.debug("ONNX Checker Passed")
        except onnx.checker.ValidationError as err:
            logging.warning("ONNX Checker exited with an error: {:}".format(err))
        return model


# ONNX loaders return ONNX models in memory.
class OnnxFileLoader(BaseOnnxModelLoader):
    def __init__(self, path):
        """
        Loads an ONNX model from a file.

        Args:
            path (str): The path from which to load the model.
        """
        self.path = path

    def __call__(self):
        logging.info("Loading {:}".format(self.path))
        return self.check(onnx.load(self.path))

    def __str__(self):
        return "ONNX Model Loader: {:}".format(self.path)

    def __repr__(self):
        return self.__str__()


class OnnxNetworkLoader(BaseModelLoader):
    def __init__(self, onnx_loader, explicit_precision=None):
        """
        Parses an ONNX model to create an engine.

        Args:
            onnx_loader (Callable() -> onnx.ModelProto): A loader that can supply an ONNX model.

        Optional Args:
            explicit_precision (bool): Whether to create the network with explicit precision enabled.
        """
        self.onnx_loader = onnx_loader
        self.explicit_precision = default_value(explicit_precision, False)

    def __call__(self):
        network = TensorRTRunnerV2.create_network(explicit_precision=self.explicit_precision)

        parser = trt.OnnxParser(network, TRT_LOGGER)
        success = parser.parse(self.onnx_loader().SerializeToString())
        if not success:
            for index in range(parser.num_errors):
                logging.error(parser.get_error(index))
            logging.critical("Could not parse ONNX correctly")

        return network, parser


class BuildEngineLoader(BaseModelLoader):
    def __init__(
        self,
        network_loader,
        max_workspace_size=None,
        fp16_mode=None,
        int8_mode=None,
        profile_shapes=None,
        write_engine=None,
        calibrator=None,
        preprocess_network=None,
        layerwise=None,
    ):
        """
        Uses a TensorRT INetworkDefinition to build an engine

        Args:
            network_loader (Callable()->trt.INetworkDefinition): A callable capable of returning an TensorRT INetworkDefinition. The returned network is owned by the BuildEngineLoader and should not be freed manually. The callable may have at most 2 return values if another object needs to be kept alive for the duration of the network, e.g., in the case of a parser. BuildEngineLoader will take ownership of the second return value, and, like the network, it should not be freed by the callable. The first return value must always be the network.

        Optional Args:
            max_workspace_size (int): The maximum workspace size, in bytes, when building the engine.
            fp16_mode (bool): Whether to build the engine with fp16 mode enabled.
            int8_mode (bool): Whether to build the engine with int8 mode enabled.
            profile_shapes (Dict[str, List[shape, shape, shape]]): A mapping of binding name to min/opt/max shapes. Only needed for networks with dynamic input shapes.
            write_engine (str): A directory in which to save the engine.
            calibrator (trt_smeagol.runners.tensorrt_runner_v2.Calibrator): An int8 calibrator. Only required in int8 mode when the network does not have explicit precision.
            preprocess_network (Callable(trt.INetworkDefinition)): Preprocessing function for the network definition. May be used to modify the network after parsing. This is called before enabling layerwise outputs.
            layerwise (bool): Whether to treat the output of every layer as an output of the network. Defaults to False.
        """
        self.network_loader = network_loader
        self.max_workspace_size = default_value(max_workspace_size, 1 << 24)
        self.fp16_mode = default_value(fp16_mode, False)
        self.int8_mode = default_value(int8_mode, False)
        self.profile_shapes = default_value(profile_shapes, OrderedDict())
        self.write_engine = write_engine
        self.written_engine_path = None
        self.calibrator = calibrator
        self.preprocess_network = default_value(preprocess_network, None)
        self.layerwise = default_value(layerwise, False)

    def __call__(self):
        class DummyContextManager(object):
            def __enter__(self):
                return None

            def __exit__(self, exc_type, exc_value, traceback):
                return None

        network_parser = self.network_loader()
        try:
            network, parser = network_parser
            assert isinstance(network, trt.INetworkDefinition)
        except (ValueError, AssertionError):
            network = network_parser
            parser = DummyContextManager()

        with trt.Builder(TRT_LOGGER) as builder, network, parser:
            if self.preprocess_network:
                logging.debug("Applying network preprocessing: {:}".format(self.preprocess_network))
                self.preprocess_network(network)

            if self.layerwise:
                TensorRTRunnerV2.mark_layerwise(network)

            if logging.getEffectiveLevel() <= logging.DEBUG:
                TensorRTRunnerV2.log_network(network)

            config = builder.create_builder_config()
            profile = TensorRTRunnerV2.build_profile(builder, network, self.profile_shapes)
            config.add_optimization_profile(profile)

            config.max_workspace_size = int(self.max_workspace_size)
            if self.fp16_mode:
                config.flags = 1 << int(trt.BuilderFlag.FP16)
            if self.int8_mode:
                config.flags = config.flags | 1 << int(trt.BuilderFlag.INT8)
                if not network.has_explicit_precision:
                    if not self.calibrator:
                        logging.critical(
                            "Network does not have explicit precision. A calibrator must be provided in order to use int8 mode."
                        )
                    self.calibrator.set_input_metadata(get_input_metadata_from_profile(profile, network))
                    config.int8_calibrator = self.calibrator

            logging.debug("Using builder configuration flags: {:}".format(config.flags))
            logging.info(
                "Building engine: max workspace size={:} bytes, fp16={:}, int8={:}, layerwise={:}".format(
                    self.max_workspace_size, self.fp16_mode, self.int8_mode, self.layerwise
                )
            )
            engine = builder.build_engine(network, config)
            self.written_engine_path = write_timestamped(
                contents=lambda: engine.serialize(), dir=self.write_engine, name="tensorrt_runner_v2.engine"
            )
            return engine

    def get_engine_path(self):
        """
        Returns the path at which the engine was written, or None if write_engine was not specified.
        """
        return self.written_engine_path



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "ActionCallback",
    "ModuleSaverCallback",
    "SimpleLossLoggerCallback",
    "EvaluatorCallback",
    "ValueSetterCallback",
    "UnfreezeCallback",
    "WandbCallback",
]

import datetime
import glob
import os
import sys
import time
from abc import ABC, abstractmethod
from collections import namedtuple

from nemo.utils import logging
from nemo.utils.decorators import deprecated

try:
    import wandb

    _WANDB_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    _WANDB_AVAILABLE = False


class ActionCallback(ABC):
    """Abstract interface for callbacks.
    """

    def __init__(self):
        self._registered_tensors = {}
        self._action = None

    @property
    def step(self):
        return self.action.step

    @property
    def epoch_num(self):
        return self.action.epoch_num

    @property
    def registered_tensors(self):
        return self._registered_tensors

    @property
    def local_rank(self):
        return self.action.local_rank

    @property
    def global_rank(self):
        return self.action.global_rank

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, action_obj):
        self._action = action_obj

    @property
    def logger(self):
        return logging

    def on_action_start(self):
        pass

    def on_action_end(self):
        pass

    def on_epoch_start(self):
        pass

    def on_epoch_end(self):
        pass

    def on_iteration_start(self):
        pass

    def on_iteration_end(self):
        pass


class ModuleSaverCallback(ActionCallback):
    """
    For callback documentation: please see
    https://nvidia.github.io/NeMo/tutorials/callbacks.html
    """

    @deprecated(version="0.12", explanation="The callback section of NeMo has been updated.")
    def __init__(
        self, save_modules_list, step_freq=1000, folder=None, checkpoints_to_keep=4,
    ):
        super().__init__()
        self._save_modules_list = save_modules_list
        self._folder = folder
        self._step_freq = step_freq
        self._ckpt2keep = checkpoints_to_keep
        self._saved_ckpts = []

    def on_iteration_end(self):
        step = self.step
        if (
            self._step_freq > 0
            and step % self._step_freq == 0
            and step > 0
            and (self.global_rank is None or self.global_rank == 0)
        ):
            for m in self._save_modules_list:
                class_name = m.__class__.__name__
                uid = m.unique_instance_id
                fn = f"{class_name}_{uid}-STEP-{step}.pt"
                if self._folder is None:
                    file_name = fn
                else:
                    file_name = os.path.join(self._folder, fn)
                logging.info(f"Saving module {class_name} in {file_name}")
                m.save_to(file_name)
                logging.info("Saved.")
            self._saved_ckpts.append(f'-{self.step}.pt')
            if len(self._saved_ckpts) > self._ckpt2keep:
                for end in self._saved_ckpts[: -self._ckpt2keep]:
                    for file in glob.glob(f'{self._folder}/*{end}'):
                        os.remove(file)
                self._saved_ckpts = self._saved_ckpts[-self._ckpt2keep :]

    def on_action_end(self):
        step = self.step
        if self.global_rank is None or self.global_rank == 0:
            for m in self._save_modules_list:
                class_name = m.__class__.__name__
                uid = m.unique_instance_id
                fn = f"{class_name}_{uid}-STEP-{step}.pt"
                if self._folder is None:
                    file_name = fn
                else:
                    file_name = os.path.join(self._folder, fn)
                logging.info(f"Saving module {class_name} in {file_name}")
                m.save_to(file_name)
                logging.info("Saved.")


class SimpleLossLoggerCallback(ActionCallback):
    """
    For callback documentation: please see
    https://nvidia.github.io/NeMo/tutorials/callbacks.html
    """

    @deprecated(version="0.12", explanation="The callback section of NeMo has been updated.")
    def __init__(
        self, tensors, print_func=None, get_tb_values=None, log_to_tb_func=None, step_freq=25, tb_writer=None,
    ):

        super().__init__()
        if not isinstance(tensors, list):
            tensors = [tensors]
        self._tensors = tensors
        self._print_func = print_func
        self._get_tb_values = get_tb_values
        self._log_to_tb_func = log_to_tb_func
        self._step_freq = step_freq
        self._swriter = tb_writer
        self._start_time = None
        self._last_epoch_start = None
        self._last_iter_start = None

    @property
    def tensors(self):
        return self._tensors

    def on_action_start(self):
        if self.global_rank is None or self.global_rank == 0:
            logging.info("Starting .....")
            self._start_time = time.time()

    def on_action_end(self):
        if self.global_rank is None or self.global_rank == 0:
            if self._swriter is not None:
                self._swriter.close()
            delta = datetime.timedelta(seconds=(time.time() - self._start_time))
            logging.info("Done in %s", delta)

    def on_epoch_start(self):
        if self.global_rank is None or self.global_rank == 0:
            logging.info(f"Starting epoch {self.epoch_num}")
            self._last_epoch_start = time.time()

    def on_epoch_end(self):
        if self.global_rank is None or self.global_rank == 0:
            step = self.step

            delta = datetime.timedelta(seconds=(time.time() - self._last_epoch_start))
            logging.info(f"Finished epoch {self.epoch_num} in {delta}")

            if self._swriter is not None:
                value = self.epoch_num
                self._swriter.add_scalar('misc/epoch', value, step)
                value = time.time() - self._last_epoch_start
                self._swriter.add_scalar('misc/epoch_time', value, step)

    def on_iteration_start(self):
        if self.global_rank is None or self.global_rank == 0:
            self._last_iter_start = time.time()

    def on_iteration_end(self):
        if self.global_rank is None or self.global_rank == 0:
            step = self.step
            if step % self._step_freq == 0:
                tensor_values = [self.registered_tensors[t.unique_name] for t in self.tensors]
                logging.info(f"Step: {step}")
                if self._print_func:
                    self._print_func(tensor_values)
                sys.stdout.flush()
                if self._swriter is not None:
                    if self._get_tb_values:
                        tb_objects = self._get_tb_values(tensor_values)
                        for name, value in tb_objects:
                            value = value.item()
                            self._swriter.add_scalar(name, value, step)
                    if self._log_to_tb_func:
                        self._log_to_tb_func(self._swriter, tensor_values, step)
                    run_time = time.time() - self._last_iter_start
                    self._swriter.add_scalar('misc/step_time', run_time, step)
                run_time = time.time() - self._last_iter_start
                logging.info(f"Step time: {run_time} seconds")

                # To keep support in line with the removal of learning rate logging from inside actions, log learning
                # rate to tensorboard. However it now logs ever self._step_freq as opposed to every step
                if self._swriter is not None:
                    self._swriter.add_scalar('param/lr', self.learning_rate, step)


class EvaluatorCallback(ActionCallback):
    """
    For callback documentation: please see
    https://nvidia.github.io/NeMo/tutorials/callbacks.html
    """

    def __init__(
        self,
        eval_tensors,
        user_iter_callback,
        user_epochs_done_callback,
        tb_writer=None,
        tb_writer_func=None,
        eval_step=1,
        eval_epoch=None,
        wandb_name=None,
        wandb_project=None,
        eval_at_start=True,
    ):
        # TODO: Eval_epoch currently does nothing
        if eval_step is None and eval_epoch is None:
            raise ValueError("Either eval_step or eval_epoch must be set. " f"But got: {eval_step} and {eval_epoch}")
        if (eval_step is not None and eval_step <= 0) or (eval_epoch is not None and eval_epoch <= 0):
            raise ValueError(f"Eval_step and eval_epoch must be > 0." f"But got: {eval_step} and {eval_epoch}")
        super().__init__()
        self._eval_tensors = eval_tensors
        self._swriter = tb_writer
        self._tb_writer_func = tb_writer_func
        self._eval_frequency = eval_step
        self._eval_at_start = eval_at_start
        # will be passed to callbacks below
        self._global_var_dict = {}

        # Callbacks
        self.user_iter_callback = user_iter_callback
        self.user_done_callback = user_epochs_done_callback

        # Weights and biases
        self._wandb_project = wandb_project
        self._wandb_name = wandb_name

    @property
    def eval_tensors(self):
        return self._eval_tensors

    @property
    def tb_writer_func(self):
        return self._tb_writer_func

    @property
    def swriter(self):
        return self._swriter

    def on_epoch_end(self):
        pass

    def on_iteration_end(self):
        if self.step == 0 and not self._eval_at_start:
            return
        if self.step % self._eval_frequency == 0:
            if self.global_rank == 0 or self.global_rank is None:
                logging.info('Doing Evaluation ' + '.' * 30)
            start_time = time.time()
            self.action._eval(self._eval_tensors, self, self.step)
            elapsed_time = time.time() - start_time
            if self.global_rank == 0 or self.global_rank is None:
                logging.info(f'Evaluation time: {elapsed_time} seconds')

    def on_action_start(self):
        if self.global_rank is None or self.global_rank == 0:
            if self._wandb_name is not None or self._wandb_project is not None:
                if _WANDB_AVAILABLE and wandb.run is None:
                    wandb.init(name=self._wandb_name, project=self._wandb_project)
                elif _WANDB_AVAILABLE and wandb.run is not None:
                    logging.info("Re-using wandb session")
                else:
                    logging.error("Could not import wandb. Did you install it (pip install --upgrade wandb)?")
                    logging.info("Will not log data to weights and biases.")
                    self._wandb_name = None
                    self._wandb_project = None

    def on_action_end(self):
        step = self.step
        if self.global_rank == 0 or self.global_rank is None:
            logging.info('Final Evaluation ' + '.' * 30)
        start_time = time.time()
        self.action._eval(self._eval_tensors, self, step)
        elapsed_time = time.time() - start_time
        if self.global_rank == 0 or self.global_rank is None:
            logging.info(f'Evaluation time: {elapsed_time} seconds')

    def clear_global_var_dict(self):
        self._global_var_dict = {}

    def wandb_log(self, tensors_logged):
        if self._wandb_name is not None and _WANDB_AVAILABLE:
            wandb.log(tensors_logged, step=self.step)


_Policy = namedtuple('Policy', 'method start end')


class _Method(ABC):
    """ Classes inherited from _Method are used for
    ValueSetterCallback below
    """

    @abstractmethod
    def __call__(self, step, total_steps):
        pass


class _Const(_Method):
    def __init__(self, value):
        super().__init__()

        self.value = value

    def __call__(self, step, total_steps):
        return self.value


class _Linear(_Method):
    def __init__(self, a, b):
        super().__init__()
        self.a, self.b = a, b

    def __call__(self, step, total_steps):
        return self.a + (step / (total_steps - 1)) * (self.b - self.a)


_Method.Const = _Const
_Method.Linear = _Linear


class ValueSetterCallback(ActionCallback):
    Policy = _Policy
    Method = _Method

    @deprecated(version="0.12", explanation="The callback section of NeMo has been updated.")
    def __init__(self, module, arg_name, policies=None, total_steps=None, tb_writer=None):
        super().__init__()

        if policies is None:
            initial_value = getattr(module, arg_name)
            policies = [_Policy(method=Const(initial_value), start=0.0, end=1.0)]

        new_policies = []
        for p in policies:
            start, end = p.start, p.end
            if isinstance(start, float):
                start = int(start * total_steps)
            if isinstance(end, float):
                end = int(end * total_steps)
            new_policies.append(_Policy(p.method, start, end))
        policies = new_policies
        assert policies[0].start == 0
        assert policies[-1].end == total_steps

        self.module = module
        self.arg_name = arg_name
        self.policies = policies
        self.total_steps = total_steps
        self.tb_writer = tb_writer

        self.cur_i = 0

    def on_iteration_start(self):
        cur_policy = self.policies[self.cur_i]
        if self.step < cur_policy.end:
            step = self.step - cur_policy.start
            total_steps = cur_policy.end - cur_policy.start
            value = cur_policy.method(step, total_steps)
            setattr(self.module, self.arg_name, value)
            if self.tb_writer is not None:
                class_name = self.module.__class__.__name__
                name = f"param/{class_name}.{self.arg_name}"
                self.tb_writer.add_scalar(name, value, self.step)
        else:
            self.cur_i += 1
            self.on_iteration_start()


class UnfreezeCallback(ActionCallback):
    @deprecated(version="0.12", explanation="The callback section of NeMo has been updated.")
    def __init__(self, modules, start_epoch=0):
        super().__init__()

        self.modules = modules
        self.start_epoch = start_epoch

    def on_iteration_start(self):
        if self.epoch_num == self.start_epoch:
            for m in self.modules:
                m.unfreeze()


class WandbCallback(ActionCallback):
    """
    Log metrics to [Weights & Biases](https://docs.wandb.com/)
    """

    @deprecated(version="0.12", explanation="The callback section of NeMo has been updated.")
    def __init__(
        self, train_tensors=[], wandb_name=None, wandb_project=None, args=None, update_freq=25,
    ):
        """
        Args:
            train_tensors: list of tensors to evaluate and log based on training batches
            wandb_name: wandb experiment name
            wandb_project: wandb project name
            args: argparse flags - will be logged as hyperparameters
            update_freq: frequency with which to log updates
        """
        super().__init__()

        if not _WANDB_AVAILABLE:
            logging.error("Could not import wandb. Did you install it (pip install --upgrade wandb)?")

        self._update_freq = update_freq
        self._train_tensors = train_tensors
        self._name = wandb_name
        self._project = wandb_project
        self._args = args

    def on_action_start(self):
        if self.global_rank is None or self.global_rank == 0:
            if _WANDB_AVAILABLE and wandb.run is None:
                wandb.init(name=self._name, project=self._project)
                if self._args is not None:
                    logging.info('init wandb session and append args')
                    wandb.config.update(self._args)
            elif _WANDB_AVAILABLE and wandb.run is not None:
                logging.info("Re-using wandb session")
            else:
                logging.error("Could not import wandb. Did you install it (pip install --upgrade wandb)?")
                logging.info("Will not log data to weights and biases.")
                self._update_freq = -1

    def on_iteration_end(self):
        # log training metrics
        if self.global_rank is None or self.global_rank == 0:
            if self.step % self._update_freq == 0 and self._update_freq > 0:
                tensors_logged = {t.name: self.registered_tensors[t.unique_name].cpu() for t in self._train_tensors}
                # Always log learning rate
                tensors_logged['LR'] = self.learning_rate
                self.wandb_log(tensors_logged)

    def on_epoch_start(self):
        if self.global_rank is None or self.global_rank == 0:
            self._last_epoch_start = time.time()

    def on_epoch_end(self):
        if self.global_rank is None or self.global_rank == 0:
            # always log epoch num and epoch_time
            epoch_time = time.time() - self._last_epoch_start
            self.wandb_log({"epoch": self.epoch_num, "epoch_time": epoch_time})

    def wandb_log(self, tensors_logged):
        if _WANDB_AVAILABLE:
            wandb.log(tensors_logged, step=self.step)



# Copyright (c) 2019 NVIDIA Corporation
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from nemo.backends.pytorch.nm import TrainableNM
from nemo.collections.asr.parts.jasper import JasperBlock, StatsPoolLayer, init_weights, jasper_activations
from nemo.core import DeviceType
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs


class JasperEncoder(TrainableNM):
    """
    Jasper Encoder creates the pre-processing (prologue), Jasper convolution
    block, and the first 3 post-processing (epilogue) layers as described in
    Jasper (https://arxiv.org/abs/1904.03288)

    Args:
        jasper (list): A list of dictionaries. Each element in the list
            represents the configuration of one Jasper Block. Each element
            should contain::

                {
                    # Required parameters
                    'filters' (int) # Number of output channels,
                    'repeat' (int) # Number of sub-blocks,
                    'kernel' (int) # Size of conv kernel,
                    'stride' (int) # Conv stride
                    'dilation' (int) # Conv dilation
                    'dropout' (float) # Dropout probability
                    'residual' (bool) # Whether to use residual or not.
                    # Optional parameters
                    'residual_dense' (bool) # Whether to use Dense Residuals
                        # or not. 'residual' must be True for 'residual_dense'
                        # to be enabled.
                        # Defaults to False.
                    'separable' (bool) # Whether to use separable convolutions.
                        # Defaults to False
                    'groups' (int) # Number of groups in each conv layer.
                        # Defaults to 1
                    'heads' (int) # Sharing of separable filters
                        # Defaults to -1
                    'tied' (bool)  # Whether to use the same weights for all
                        # sub-blocks.
                        # Defaults to False
                    'se' (bool)  # Whether to add Squeeze and Excitation
                        # sub-blocks.
                        # Defaults to False
                    'se_reduction_ratio' (int)  # The reduction ratio of the Squeeze
                        # sub-module.
                        # Must be an integer > 1.
                        # Defaults to 8.
                    'se_context_window' (int) # The size of the temporal context
                        # provided to SE sub-module.
                        # Must be an integer. If value <= 0, will perform global
                        # temporal pooling (global context).
                        # If value >= 1, will perform stride 1 average pooling to
                        # compute context window.
                    'se_interpolation_mode' (str) # Interpolation mode of timestep dimension.
                        # Used only if context window is > 1.
                        # The modes available for resizing are: `nearest`, `linear` (3D-only),
                        # `bilinear`, `area`
                    'kernel_size_factor' (float)  # Conv kernel size multiplier
                        # Can be either an int or float
                        # Kernel size is recomputed as below:
                        # new_kernel_size = int(max(1, (kernel_size * kernel_width)))
                        # to prevent kernel sizes than 1.
                        # Note: If rescaled kernel size is an even integer,
                        # adds 1 to the rescaled kernel size to allow "same"
                        # padding.
                    'stride_last' (bool) # Bool flag to determine whether each
                        # of the the repeated sub-blockss will perform a stride,
                        # or only the last sub-block will perform a strided convolution.
                }

        activation (str): Activation function used for each sub-blocks. Can be
            one of ["hardtanh", "relu", "selu", "swish"].
        feat_in (int): Number of channels being input to this module
        normalization_mode (str): Normalization to be used in each sub-block.
            Can be one of ["batch", "layer", "instance", "group"]
            Defaults to "batch".
        residual_mode (str): Type of residual connection.
            Can be "add", "stride_add" or "max".
            "stride_add" mode performs strided convolution prior to residual
            addition.
            Defaults to "add".
        norm_groups (int): Number of groups for "group" normalization type.
            If set to -1, number of channels is used.
            Defaults to -1.
        conv_mask (bool): Controls the use of sequence length masking prior
            to convolutions.
            Defaults to True.
        frame_splicing (int): Defaults to 1.
        init_mode (str): Describes how neural network parameters are
            initialized. Options are ['xavier_uniform', 'xavier_normal',
            'kaiming_uniform','kaiming_normal'].
            Defaults to "xavier_uniform".
    """

    length: Optional[torch.Tensor]

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "audio_signal": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "audio_signal": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "outputs": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(EncodedRepresentationTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "encoded_lengths": NeuralType({0: AxisType(BatchTag)}),
            "outputs": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation()),
            "encoded_lengths": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def _disabled_deployment_input_ports(self):
        return set(["length"])

    @property
    def _disabled_deployment_output_ports(self):
        return set(["encoded_lengths"])

    def _prepare_for_deployment(self):
        m_count = 0
        for m in self.modules():
            if type(m).__name__ == "MaskedConv1d":
                m.use_mask = False
                m_count += 1
        logging.warning(f"Turned off {m_count} masked convolutions")
        input_example = torch.randn(16, self.__feat_in, 256)
        if self._placement == DeviceType.GPU or self._placement == DeviceType.AllGpu:
            input_example = input_example.cuda()
        return input_example, None

    def __init__(
        self,
        jasper,
        activation,
        feat_in,
        normalization_mode="batch",
        residual_mode="add",
        norm_groups=-1,
        conv_mask=True,
        frame_splicing=1,
        init_mode='xavier_uniform',
    ):
        super().__init__()

        activation = jasper_activations[activation]()
        feat_in = feat_in * frame_splicing

        self.__feat_in = feat_in

        residual_panes = []
        encoder_layers = []
        self.dense_residual = False
        for lcfg in jasper:
            dense_res = []
            if lcfg.get('residual_dense', False):
                residual_panes.append(feat_in)
                dense_res = residual_panes
                self.dense_residual = True
            groups = lcfg.get('groups', 1)
            separable = lcfg.get('separable', False)
            heads = lcfg.get('heads', -1)
            residual_mode = lcfg.get('residual_mode', residual_mode)
            se = lcfg.get('se', False)
            se_reduction_ratio = lcfg.get('se_reduction_ratio', 8)
            se_context_window = lcfg.get('se_context_window', -1)
            se_interpolation_mode = lcfg.get('se_interpolation_mode', 'nearest')
            kernel_size_factor = lcfg.get('kernel_size_factor', 1.0)
            stride_last = lcfg.get('stride_last', False)
            encoder_layers.append(
                JasperBlock(
                    feat_in,
                    lcfg['filters'],
                    repeat=lcfg['repeat'],
                    kernel_size=lcfg['kernel'],
                    stride=lcfg['stride'],
                    dilation=lcfg['dilation'],
                    dropout=lcfg['dropout'],
                    residual=lcfg['residual'],
                    groups=groups,
                    separable=separable,
                    heads=heads,
                    residual_mode=residual_mode,
                    normalization=normalization_mode,
                    norm_groups=norm_groups,
                    activation=activation,
                    residual_panes=dense_res,
                    conv_mask=conv_mask,
                    se=se,
                    se_reduction_ratio=se_reduction_ratio,
                    se_context_window=se_context_window,
                    se_interpolation_mode=se_interpolation_mode,
                    kernel_size_factor=kernel_size_factor,
                    stride_last=stride_last,
                )
            )
            feat_in = lcfg['filters']

        self.encoder = nn.Sequential(*encoder_layers)
        self.apply(lambda x: init_weights(x, mode=init_mode))
        self.to(self._device)

    def forward(self, audio_signal, length=None):
        # type: (Tensor, Optional[Tensor]) -> Tensor, Optional[Tensor]

        s_input, length = self.encoder(([audio_signal], length))
        if length is None:
            return s_input[-1]

        return s_input[-1], length


class JasperDecoderForCTC(TrainableNM):
    """
    Jasper Decoder creates the final layer in Jasper that maps from the outputs
    of Jasper Encoder to the vocabulary of interest.

    Args:
        feat_in (int): Number of channels being input to this module
        num_classes (int): Number of characters in ASR model's vocab/labels.
            This count should not include the CTC blank symbol.
        init_mode (str): Describes how neural network parameters are
            initialized. Options are ['xavier_uniform', 'xavier_normal',
            'kaiming_uniform','kaiming_normal'].
            Defaults to "xavier_uniform".
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "encoder_output": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(EncodedRepresentationTag), 2: AxisType(ProcessedTimeTag),}
            # )
            "encoder_output": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"output": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(ChannelTag),})}
        return {"output": NeuralType(('B', 'T', 'D'), LogprobsType())}

    def __init__(self, feat_in, num_classes, init_mode="xavier_uniform", vocabulary=None):
        if vocabulary is not None:
            if num_classes != len(vocabulary):
                raise ValueError(
                    f"If vocabulary is specified, it's length should be equal to the num_classes. But I got: num_classes={num_classes} and len(vocabluary)={len(vocabulary)}"
                )
            self.__vocabulary = vocabulary

        super().__init__()
        self._feat_in = feat_in
        # Add 1 for blank char
        self._num_classes = num_classes + 1

        self.decoder_layers = nn.Sequential(nn.Conv1d(self._feat_in, self._num_classes, kernel_size=1, bias=True))
        self.apply(lambda x: init_weights(x, mode=init_mode))
        self.to(self._device)

    def forward(self, encoder_output):
        return F.log_softmax(self.decoder_layers(encoder_output).transpose(1, 2), dim=-1)

    def _prepare_for_deployment(self):
        input_example = torch.randn(34, self._feat_in, 1)
        if self._placement == DeviceType.GPU or self._placement == DeviceType.AllGpu:
            input_example = input_example.cuda()
        return input_example, None

    @property
    def vocabulary(self):
        return self.__vocabulary


class JasperDecoderForClassification(TrainableNM):
    """
        Jasper Decoder creates the final layer in Jasper that maps from the outputs
        of Jasper Encoder to one class label.

        Args:
            feat_in (int): Number of channels being input to this module
            num_classes (int): Number of characters in ASR model's vocab/labels.
                This count should not include the CTC blank symbol.
            init_mode (str): Describes how neural network parameters are
                initialized. Options are ['xavier_uniform', 'xavier_normal',
                'kaiming_uniform','kaiming_normal'].
                Defaults to "xavier_uniform".
        """

    @property
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "encoder_output": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(EncodedRepresentationTag), 2: AxisType(ProcessedTimeTag)}
            # )
            "encoder_output": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation())
        }

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"logits": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag)})}
        return {"logits": NeuralType(('B', 'D'), LogitsType())}

    def __init__(
        self, *, feat_in, num_classes, init_mode="xavier_uniform", return_logits=True, pooling_type='avg', **kwargs
    ):
        TrainableNM.__init__(self, **kwargs)

        self._feat_in = feat_in
        self._return_logits = return_logits
        self._num_classes = num_classes

        if pooling_type == 'avg':
            self.pooling = nn.AdaptiveAvgPool1d(1)
        elif pooling_type == 'max':
            self.pooling = nn.AdaptiveMaxPool1d(1)
        else:
            raise ValueError('Pooling type chosen is not valid. Must be either `avg` or `max`')

        self.decoder_layers = nn.Sequential(nn.Linear(self._feat_in, self._num_classes, bias=True))
        self.apply(lambda x: init_weights(x, mode=init_mode))
        self.to(self._device)

    def forward(self, encoder_output):
        batch, in_channels, timesteps = encoder_output.size()

        encoder_output = self.pooling(encoder_output).view(batch, in_channels)  # [B, C]
        logits = self.decoder_layers(encoder_output)  # [B, num_classes]

        if self._return_logits:
            return logits

        return F.softmax(logits, dim=-1)


class JasperDecoderForSpkrClass(TrainableNM):
    """
    Jasper Decoder creates the final layer in Jasper that maps from the outputs
    of Jasper Encoder to the embedding layer followed by speaker based softmax loss.

    Args:
        feat_in (int): Number of channels being input to this module
        num_classes (int): Number of unique speakers in dataset
        emb_sizes (list) : shapes of intermediate embedding layers (we consider speaker embbeddings from 1st of this layers)
                Defaults to [1024,1024]
        pool_mode (str) : Pooling stratergy type. options are 'gram','xvector','superVector'.
                Defaults to 'xvector'
        init_mode (str): Describes how neural network parameters are
            initialized. Options are ['xavier_uniform', 'xavier_normal',
            'kaiming_uniform','kaiming_normal'].
            Defaults to "xavier_uniform".
    """

    @property
    def input_ports(self):
        """Returns definitions of module input ports.

        encoder_output:
            0: AxisType(BatchTag)

            1: AxisType(EncodedRepresentationTag)

            2: AxisType(ProcessedTimeTag)
        """

        return {"encoder_output": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation())}

    @property
    def output_ports(self):
        """Returns definitions of module output ports.

        logits:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag)
                
        embs: 
            0: AxisType(BatchTag)
            1: AxisType(EncodedRepresentationTah) 
        """
        return {
            "logits": NeuralType(('B', 'D'), LogitsType()),
            "embs": NeuralType(('B', 'D'), AcousticEncodedRepresentation()),
        }

    def __init__(self, feat_in, num_classes, emb_sizes=[1024, 1024], pool_mode='xvector', init_mode="xavier_uniform"):
        TrainableNM.__init__(self)
        self._feat_in = 0
        if pool_mode == 'gram':
            gram = True
            super_vector = False
        elif pool_mode == 'superVector':
            gram = True
            super_vector = True
        else:
            gram = False
            super_vector = False

        if gram:
            self._feat_in += feat_in ** 2
        else:
            self._feat_in += 2 * feat_in

        if super_vector and gram:
            self._feat_in += 2 * feat_in

        self._midEmbd1 = int(emb_sizes[0])  # Spkr Vector Embedding Shape
        self._midEmbd2 = int(emb_sizes[1]) if len(emb_sizes) > 1 else 0  # Spkr Vector Embedding Shape

        self._num_classes = num_classes
        self._pooling = StatsPoolLayer(gram=gram, super_vector=super_vector)

        self.mid1 = self.affineLayer(self._feat_in, self._midEmbd1, learn_mean=False)
        self.mid2 = self.affineLayer(self._midEmbd1, self._midEmbd2, learn_mean=False)
        self.final = nn.Linear(self._midEmbd2, self._num_classes)

        self.apply(lambda x: init_weights(x, mode=init_mode))
        self.to(self._device)

    def affineLayer(self, inp_shape, out_shape, learn_mean=True):
        layer = nn.Sequential(
            nn.Linear(inp_shape, out_shape),
            nn.BatchNorm1d(out_shape, affine=learn_mean, track_running_stats=True),
            nn.ReLU(),
        )

        return layer  # layer, embs

    def forward(self, encoder_output):
        # encoder_output = self.norm(encoder_output)
        pool = self._pooling(encoder_output)
        mid1, emb1 = self.mid1(pool), self.mid1[:2](pool)
        mid2, embs = self.mid2(mid1), self.mid2[:2](mid1)
        out = self.final(mid2)

        return out, emb1


# Siamese Network, support to be added in future releases
# class SiameseDecoderForSpeakerClass(TrainableNM):
#     """
#     Jasper Decoder creates the final layer in Jasper that maps from the outputs
#     of Jasper Encoder to the vocabulary of interest.

#     Args:
#         feat_in (int): Number of channels being input to this module
#         num_classes (int): Number of characters in ASR model's vocab/labels.
#             This count should not include the CTC blank symbol.
#         init_mode (str): Describes how neural network parameters are
#             initialized. Options are ['xavier_uniform', 'xavier_normal',
#             'kaiming_uniform','kaiming_normal'].
#             Defaults to "xavier_uniform".
#     """

#     @property
#     def input_ports(self):
#         """Returns definitions of module input ports.

#         encoder_output:
#             0: AxisType(BatchTag)

#             1: AxisType(EncodedRepresentationTag)

#             2: AxisType(ProcessedTimeTag)
#         """
#         return {
#             "embs1": NeuralType(('B', 'D'), AcousticEncodedRepresentation()),
#             "embs2": NeuralType(('B', 'D'), AcousticEncodedRepresentation()),
#         }

#     @property
#     def output_ports(self):
#         """Returns definitions of module output ports.

#         output:
#             0: AxisType(BatchTag)

#             1: AxisType(ChannelTag)
#         """
#         return {
#             "logits": NeuralType(('B', 'D'), LogitsType()),
#         }

#     def __init__(self, emb_size, mid_dim, init_mode="xavier_uniform"):
#         super().__init__()
#         self._feat_in = emb_size
#         self._mid_dim = mid_dim

#         self.connect = self.affineLayer(self._feat_in, self._mid_dim, learn_mean=True)

#         self.S = nn.Parameter(torch.randn(self._mid_dim, self._mid_dim), requires_grad=True)
#         self.b = nn.Parameter(torch.randn(1), requires_grad=True)

#         self.apply(lambda x: init_weights(x, mode=init_mode))
#         self.to(self._device)

#     def affineLayer(self, inp_shape, out_shape, learn_mean=True):
#         layer = nn.Sequential(
#             nn.Linear(inp_shape, out_shape),
#             nn.BatchNorm1d(out_shape, affine=learn_mean, track_running_stats=True),
#             nn.ReLU(),
#         )

#         return layer  # layer, embs

#     def forward(self, inp_emb1, inp_emb2):

#         x = self.connect(inp_emb1)
#         y = self.connect(inp_emb2)

#         out = (
#             torch.matmul(x, y.T).diag()
#             - torch.matmul(torch.matmul(x, self.S), x.T).diag()
#             - torch.matmul(torch.matmul(y, self.S), y.T).diag()
#             + self.b
#         )

#         return out



from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils import (
    fill_class_weights,
    get_freq_weights,
    get_label_stats,
    if_exist,
)

__all__ = ['TextClassificationDataDesc']


class TextClassificationDataDesc:
    def __init__(self, data_dir, modes=['train', 'test', 'dev']):
        self.data_dir = data_dir

        max_label_id = 0
        for mode in modes:
            if not if_exist(self.data_dir, [f'{mode}.tsv']):
                logging.info(f'Stats calculation for {mode} mode is skipped as {mode}.tsv was not found.')
                continue

            input_file = f'{self.data_dir}/{mode}.tsv'
            with open(input_file, 'r') as f:
                input_lines = f.readlines()[1:]  # Skipping headers at index 0

            try:
                int(input_lines[0].strip().split()[-1])
            except ValueError:
                logging.warning(f'No numerical labels found for {mode}.tsv.')
                raise

            queries, raw_sentences = [], []
            for input_line in input_lines:
                parts = input_line.strip().split()
                label = int(parts[-1])
                raw_sentences.append(label)
                queries.append(' '.join(parts[:-1]))

            infold = input_file[: input_file.rfind('/')]

            logging.info(f'Three most popular classes in {mode} dataset')
            total_sents, sent_label_freq, max_id = get_label_stats(
                raw_sentences, infold + f'/{mode}_sentence_stats.tsv'
            )
            max_label_id = max(max_label_id, max_id)

            if mode == 'train':
                class_weights_dict = get_freq_weights(sent_label_freq)
                logging.info(f'Class Weights: {class_weights_dict}')

            logging.info(f'Total Sentences: {total_sents}')
            logging.info(f'Sentence class frequencies - {sent_label_freq}')

        self.class_weights = fill_class_weights(class_weights_dict, max_label_id)

        self.num_labels = max_label_id + 1
